/*
 * concatenation.hpp
 *
 *  Created on: Aug 14, 2017
 *      Author: julien
 */

#ifndef SRC_MODULES_CONCATENATION_CONCATENATION_HPP_
#define SRC_MODULES_CONCATENATION_CONCATENATION_HPP_


#include "../../OpenCV_includes.hpp"
#include "../helper_functions/helper_functions.hpp"



class Concatenation{

	private:

		enum side{left, right};

		cv::Mat lookup_x; // Mat of the x-gradient
		cv::Mat lookup_y; // Mat of the y-gradient

		// the initial contours from the contour-function
		std::vector<std::vector<cv::Point>> initial_contours;
		// temporal std::vector for the final output of the lane-contours
		std::vector<std::vector<cv::Point>> temp_contours;
		// Point collection with any inner std::vector representing an y-value
		std::vector<std::vector<cv::Point>> structured_points_l;

		std::vector<std::vector<cv::Point>> structured_points_r;
		// this is just for the contour function and has no further use
		std::vector<cv::Vec4i> hierarchy;
		// Loose collection of all Points
		std::vector<cv::Point> contour_point_collection;

		// these are the arrays where the points are collected based upon their relative
		// position, to form lines. Both are used in the concat functions
		std::vector<cv::Point> left_contour;
		std::vector<cv::Point> right_contour;

		cv::Point translation; // for the determination of the roi-translation

		// these two arrays are for sorting out the lines that are too short
		// in concat_helper
		std::vector<cv::Point> temp_left_contour;
		std::vector<cv::Point> temp_right_contour;

		// for the filtering of the points
		// minimum x-gradient that has to be reached to be considered part of the lane marking
		unsigned short gradient_threshold;
		// how far apart do points need to be before not being deleted (on the x-axis
		int line_distance_threshold;
		// how far away does the next point be away to be considered part of the
		// same line. (on the x-axis)
		int line_spread;
		// how many points have to be in a line to consider it an actual marking
		int min_line_length;



		// half the size of the roi
//		int roi_half = 0;

//		int picture_half = 0; // half the size of the picture
//		int roi_left_area = 0;
//		int roi_right_area = 0;

		/**
		 * Collects all edges based upon their gradient-value and sorts them into
		 * structured_points_l and structured_points_r based upon their y-value
		 * and their relative position to the middle of the picture
		 * structured_points_l and structured_points_r will be sorted like so:
		 * (where p stands for Point and x, and y stand for their respective coordinates)
		 * 	[
		 *  - P(x11, y1), P(x12, y1), P(x13, y1), ... , P(x1n1, y1);
		 *  - P(x21, y2), P(x22, y2), P(x23, y2), ... , P(x2n2, y2);
		 *  - ...
		 *  - ...
		 *  - P(xm1, ym), P(xm2, ym), P(xm2, ym), ... , P(xmnm, ym);
		 *  ]
		 * the sub std::vectors don't have to be the same length as indicated by the indices.
		 */
		void collect_points();


		/**
		 * this is a helper function for collect_points()
		 * it determines, where the value of current_point needs to be saved
		 *
		 * @param std::vector<std::vector<Point>>* structured_points : A pointer to either
		 * 												     structured_points_l or
		 * 												      structured_points_l.
		 * 							 						 This is the destination where
		 * 							 						 current_point will be saved to
		 * @param Point* current_point : A pointer to a point from contour_point_collection
		 * 								 that will be sorted into structured_points
		 * @param int* current_sub_std::vector : a pointer to the index of the current sub std::vector
		 * 									of structured_points
		 * @param int* current_y : a pointer to the currently looked at y value. This is the value,
		 * 						   that current_point will be compared against
		 * 						   to determine if a new sub std::vector needs to be created.
		 * 						   (based upon the specification of sort_point_collection())
		 */
		void add_to_vec(std::vector< std::vector< cv::Point >>* structured_points,
						cv::Point* current_point,
						int* current_sub_vector,
						int* current_y);

		// TODO
		void add_nearest(std::vector<cv::Point>* contour,
				cv::Point reference_point,
				std::vector<cv::Point>* lane_contour);

		/**
		 * Finds the lane markings in structured_points_r and structured_points_l,
		 * concatenates them and puts them
		 * into two seperate sub verctors of temp_contours.
		 * The values of temp_contours will serve as output to concatenate
		 */
		void concat();

		/**
		 * merges the lines that are basically the same
		 */
		void merge();

		/**
		 * filters the useless lines
		 */
		void filter();

		/**
		 * Helper function for concat().
		 * Collects Edge Points from the bottom up and puts them in a lane_contour
		 * as a sorted collection
		 *
		 * @param std::vector<std::vector<Point>>* structured_points : structured_points_l or
		 * 													 structured_points_r
		 * @param std::vector<Point>* lane_contour : the output std::vector for the filtered Points
		 * TODO
		 */
		void concat_helper(std::vector<std::vector<cv::Point>>* structured_points,
						   std::vector<cv::Point>* lane_contour,
						   side s);

		/**
		 * sifts out all the points that are too close to each other on the x-axis.
		 * Also flips the orientation of structured_points. the y-values now go down.
		 * So the points are going from the bottom of the picture to the top.
		 *
		 * @param std::vector<std::vector<Point>>* structured_points : The Collection of Points where the
		 * 													 the closest one will be sift out of.
		 * 													 so either structured_points_l or
		 * 													 structured_points_r.
		 * @param int spread_threshold :
		 */
		void sift_out(std::vector<std::vector<cv::Point>>* structured_points, side s);


		/**
		 *  coninues the given lines with the closest points out of next_points and
		 *  adds new lines to "lines", if it seems appropriate
		 *
		 *  @param std::vector<std::vector<Point>>* lines : lines that will be continued
		 *  @param std::vector<Point>* next_points : Points that either continue the last lines,
		 *  									or start new lines, if they are far enopugh away
		 *
		 *  lines that are created by this method and deemed to small to be considered lanes
		 *  will be deleted later, but that is not the concern of this method
		 *  For that there is the mehtod TODO
		 */
		void continue_lines(std::vector<std::vector<cv::Point>>* lines, std::vector<cv::Point>* next_points);

		/**
		 * calculates the difference between Point a and Point b based upon the hypotenuse
		 * inside the src-coordinates
		 * @param Point a : Point to compare against (needs to be inside the boundaries of roi,
		 * 					but in src-coordinates)
		 * @param Point b : Point to compare against (needs to be inside the boundaries of roi,
		 * 					but in src-coordinates)
		 *
		 * @return int : returns the radial difference as an integer
		 */
		double radial_distance(cv::Point a, cv::Point b);


		/**
		 * filters out Points that are "out of place" inside the contour.
		 * The other functions (based upon their algorithms) put some far off points into
		 * the contour
		 * This functions corrects this by only leaving points that are near to each other.
		 *
		 * @param std::vector<pPoint>* contour : a collection of points on which this filter will
		 * 								    be applied.
		 * 								    This is also the output place for the function.
		 * @param int lookahead : The number of Points that will be compared against each
		 * 						  other to determine the closest.
		 * 						  lookahead has to be at least 2!
		 */
		void filtercontour(std::vector<cv::Point>* contour, int lookahead);

	public:

		/**
		 * Initializes Concatenation
		 * @param int gradient_threshold : absolute value of the minimum inward gradient,
		 * 								   for which the contours will be filtered
		 * @param int distance_threshold : how big should the distance between points on the
		 * 								   x-axis be, to be considered on the same lane contour
		 * 								   (must be positive)
		 * @param int translation_x : x - translation of the roi
		 * @param int translation_y : y - translation of the roi
		 */
		Concatenation(unsigned short gradient_threshold,
					  int line_distance_threshold,
					  int line_spread,
					  int min_line_length,
					  int translation_x,
					  int translation_y);

		/**
		 * Selects the inside of the lane markings based upon grad_x and grad_y and outputs it
		 * into contours as a two dimensional std::vector of point-std::vectors.
		 * (two dimensions for left and right lane markings)
		 * @param Mat* grad_x : gradient map for the roi in x direction should be the output
		 * 						of the Scharr-function
		 * @param Mat* grad_y : gradient map for the roi in y direction should be the output
		 * 						of the Scharr-function
		 * @param std::vector<std::vector<Point>>* contours : storage place for the output of the inside
		 * 											contours of the lane
		 */
		void concatenate(cv::Mat* grad_x,
						 cv::Mat* grad_y,
						 std::vector<cv::Point>* contours_l,
						 std::vector<cv::Point>* contours_r);
};




#endif /* SRC_MODULES_CONCATENATION_CONCATENATION_HPP_ */
